import React, { PureComponent } from 'react';
import * as PropTypes from 'prop-types';
import {

    // DataTypeProvider, // used to populate custom filter operation drop down items
    FilteringState,
    SelectionState,
    VirtualTableState,
} from '@devexpress/dx-react-grid';
import {
    DragDropProvider,
    Grid,
    TableColumnReordering,
    TableColumnResizing,
    TableFilterRow,
    TableFixedColumns,
    TableHeaderRow,
    TableSelection,
    VirtualTable,
} from '@devexpress/dx-react-grid-material-ui';

import DataTableHeaderCell from './DataTableHeaderCell';
import DataTableFilterCell, { SORT_DIRECTIONS } from './DataTableFilterCell';
import DataTableCell from './DataTableCell';
import ActivityIndicator from '../ActivityIndicator/ActivityIndicator';

import getStyles from './DataTable.styles';

const DEFAULT_COLUMN_WIDTH = 200;

/**
 * Enum for filter operations
 * @typedef {Object} FilterOperations
 * @enum {String}
 */
const FILTER_OPERATIONS = {
    CONTAINS: 'contains',
    NOT_CONTAINS: 'notContains',
    STARTS_WITH: 'startsWith',
    ENDS_WITH: 'endsWith',
    EQUAL: 'equal',
    NOT_EQUAL: 'notEqual',
    GREATER_THAN: 'greaterThan',
    GREATER_THAN_OR_EQUAL: 'greaterThanOrEqual',
    LESS_THAN: 'lessThan',
    LESS_THAN_OR_EQUAL: 'lessThanOrEqual',
};


/**
 * An extensive table grid control
 * @see https://devexpress.github.io/devextreme-reactive/react/grid/
 */
class DataTable extends PureComponent {
    constructor(props) {
        super(props);
        const {
            filterDirectives,
            selectedRowIndexes,
        } = props;
        this.state = {
            filters: filterDirectives,
            previousFilters: filterDirectives,
            selection: selectedRowIndexes,
        };
    }

    render() {
        const {
            columns, // [ { name: String, title?: String } ]
            contextStyles, // Object
            defaultColumnWidth, // Number
            fixedColumnIndexes, // [ Number ]
            infiniteScrolling, // Boolean
            loading, // Boolean
            pageSize, // Number
            rows, // [ Object ]
            showFilterRow, // Boolean
            testID, // String
            totalRows, // Number
        } = this.props;
        const {
            filters,
            selection,
        } = this.state;

        console.log('DATA TABLE RXG: RENDER: ', this.props); // eslint-disable-line
        // console.log('DATA TABLE RXG: RENDER PROPS STRING: ', JSON.stringify(props)); // eslint-disable-line
        // console.log('DATA TABLE RXG: SORT DIRECTIVES: ', sortDirectives); // eslint-disable-line
        // console.log('DATA TABLE RXG: FILTER DIRECTIVES: ', filterDirectives); // eslint-disable-line
        // console.log('DATA TABLE RXG: FILTER OPERATIONS STRING: ', JSON.stringify(filterOperations)); // eslint-disable-line
        // console.log('DATA TABLE RXG: FILTER OPERATIONS: ', filterOperations); // eslint-disable-line

        const styles = getStyles(contextStyles);
        const columnOrder = columns.map((c) => (c.name));
        const columnWidths = columns.map((c) => {
            return {
                columnName: c.name,
                width: c.width || defaultColumnWidth,
            };
        });
        const fixedLeftColumns = [];
        fixedColumnIndexes.forEach((idx) => {
            const columnName = columns[idx].name;
            fixedLeftColumns.push(columnName);
        });


        return (
            <div
                className="c-data-table__container"
                style={ styles.container }
                data-test-id={ `${testID}__data-table__container` }>
                { loading && <ActivityIndicator centered /> }
                <Grid
                    columns={ columns }
                    rows={ rows }>

                    { /* Include Custom Filtering Drop Down Operations */ }
                    { /* filterOperations.map((filterOperation) => {
                        const {
                            columnName,
                            operators,
                        } = filterOperation;
                        return (
                            <DataTypeProvider
                                key={ columnName }
                                for={ [ columnName ] }
                                availableFilterOperations={ operators } />
                        );
                    }) */ }

                    { /* Include Infinite Scrolling Functionality and Handling */ }
                    { infiniteScrolling &&
                        <VirtualTableState
                            infiniteScrolling={ infiniteScrolling }
                            loading={ loading }
                            pageSize={ pageSize }
                            totalRowCount={ totalRows }
                            skip={ 0 }
                            getRows={ this.onGetRows } />
                    }

                    { /* Include Row Selection Handling */ }
                    <SelectionState
                        selection={ selection }
                        onSelectionChange={ this.handleSelectionChange } />

                    { /* Include Filtering Functionality and Handling */ }
                    { showFilterRow &&
                        <FilteringState
                            filters={ filters }
                            onFiltersChange={ this.handleFiltersChange } />
                    }

                    { /* Include Drag and Drop Functionality and Handling */ }
                    <DragDropProvider />

                    { /* Include Table */ }
                    <VirtualTable
                        cellComponent={ this.getTableCellComponent } />

                    { /* Include Column Reordering Functionality and Handling */ }
                    { /* Controlled With Internal Grid State */ }
                    <TableColumnReordering
                        defaultOrder={ columnOrder }
                        onOrderChange={ this.handleOrderChange } />

                    { /* Include Column Resizing Functionality and Handling */ }
                    { /* Controlled With Internal Grid State */ }
                    <TableColumnResizing
                        defaultColumnWidths={ columnWidths }
                        onColumnWidthsChange={ this.handleColumnWidthsChange } />

                    { /* Include Column Header Functionality and Handling */ }
                    <TableHeaderRow
                        contentComponent={ this.getHeaderCellComponent } />

                    { /* Include Filter Row when enabled */ }
                    { showFilterRow &&
                        <TableFilterRow
                            cellComponent={ this.getFilterCellComponent }
                            rowHeight={ 145 }
                            showFilterSelector />
                    }

                    { /* Include Row Selection Functionality */ }
                    <TableSelection
                        selectByRowClick
                        highlightRow
                        showSelectionColumn={ false } />

                    { /* Include Fixed/Frozen/Pinned columns */ }
                    <TableFixedColumns
                        leftColumns={ fixedLeftColumns } />
                </Grid>
            </div>
        );
    };

    getHeaderCellComponent = (baseProps) => {
        console.log('DATA TABLE DXG: HEADER CELL: BASE PROPS: ', baseProps); // eslint-disable-line
        const {
            columns, // [ { name: String, title?: String } ]
            fixedColumnIndexes, // [ Number ]
            onColumnFrozen, // Function { String } columnName
            onFilterToggle, // Function { String } columnName
            showFilterRow, // Boolean
        } = this.props;
        const {
            // Unused available props
            // align: PropTypes.bool,
            // children: PropTypes.oneOfType([
            //     PropTypes.array,
            //     PropTypes.node,
            //     PropTypes.element,
            // ]),
            column,
        } = baseProps;
        const {
            name,
            title,
        } = column;
        const fixedColumnsMap = {};

        fixedColumnIndexes.forEach((idx) => {
            fixedColumnsMap[columns[idx].name] = idx;
        });

        return (
            <DataTableHeaderCell
                columnName={ name }
                columnTitle={ title }
                columnLocked={ fixedColumnsMap[name] !== undefined }
                onColumnFrozen={ onColumnFrozen }
                onFilterToggle={ onFilterToggle }
                showFilterRow={ showFilterRow } />
        );
    };

    getFilterCellComponent = (baseProps) => {
        const {
            onSortAscending, // Function { String } columnName
            onSortClear, // Function { String } columnName
            onSortDescending, // Function { String } columnName
            sortDirectives, // [ { columnName: String, direction: 'asc'|'desc'|null } ]
        } = this.props;
        const {
            // Unused available props
            // colSpan: PropTypes.number,
            // filter: PropTypes.object,
            // filteringEnabled: PropTypes.bool,
            // getMessage: PropTypes.func,
            // onFilter: PropTypes.func,
            // rowSpan: PropTypes.number,
            // tableRow: PropTypes.object,
            // tableColumn: PropTypes.object,
            // children: PropTypes.oneOfType([
            //     PropTypes.array,
            //     PropTypes.node,
            //     PropTypes.element,
            // ]),
            column,
        } = baseProps;
        const {
            name,
        } = column;

        const sortDirectiveIndex = sortDirectives.findIndex((sortedColumn) => (sortedColumn.columnName === name));
        let sortDirective = {};
        let sortOrder = 0;
        if (sortDirectiveIndex !== -1) {
            sortDirective = sortDirectives[sortDirectiveIndex];
            sortOrder = sortDirectiveIndex + 1;
        }
        const {
            direction,
        } = sortDirective;

        return (
            <DataTableFilterCell
                { ...baseProps }
                column={ column }
                onFilterCancel={ this.handleFilterCancel }
                onFilterReset={ this.handleFilterReset }
                onFilterSubmit={ this.handleFilterSubmit }
                onSortAscending={ onSortAscending }
                onSortClear={ onSortClear }
                onSortDescending={ onSortDescending }
                sortDirection={ direction }
                sortOrder={ sortOrder } />
        );
    };

    getTableCellComponent = (baseProps) => {
        const {
            cellComponent: TableCell, // Function
        } = this.props;

        // Unused available props
        // const {
        //     children,
        //     className,
        //     colSpan,
        //     column,
        //     row,
        //     rowSpan,
        //     style,
        //     tableColumn,
        //     tableRow,
        //     value,
        // } = baseProps;

        return (
            <TableCell { ...baseProps } />
        );
    };

    /**
     * Called on column resize
     * @param {Array.<{ columnName: String, width: Number }>} columnsWithWidths
     * @returns {void}
     */
    handleColumnWidthsChange = (columnsWithWidths) => {
        const {
            onColumnsResized,
        } = this.props;
        if (onColumnsResized) {
            onColumnsResized(columnsWithWidths);
        }
    };

    handleFilterCancel = () => {
        const {
            onFilterCancel,
        } = this.props;
        const {
            previousFilters,
        } = this.state;
        this.setState({
            filters: previousFilters,
        }, () => {
            if (onFilterCancel) {
                onFilterCancel();
            }
        });
    };

    handleFiltersChange = (filters) => {
        const {
            onFilterChange,
        } = this.props;
        this.setState({
            filters,
        }, () => {
            if (onFilterChange) {
                onFilterChange(filters);
            }
        });
    };

    handleFilterReset = () => {
        const {
            onFilterReset,
        } = this.props;
        this.setState({
            filters: [],
        }, () => {
            if (onFilterReset) {
                onFilterReset();
            }
        });
    };

    handleFilterSubmit = () => {
        const {
            onFilterSubmit,
        } = this.props;
        const {
            filters,
        } = this.state;
        this.setState({
            previousFilters: filters,
        }, () => {
            if (onFilterSubmit) {
                onFilterSubmit(filters);
            }
        });
    };

    /**
     * Called on column reorder
     * @param {Array.<String>} columnsNamesInOrder
     * @returns {void}
     */
    handleOrderChange = (columnsNamesInOrder) => {
        const {
            onColumnsReordered,
        } = this.props;
        if (onColumnsReordered) {
            onColumnsReordered(columnsNamesInOrder);
        }
    };

    handleSelectionChange = (selectedRows) => {
        const {
            onRowSelectionChange,
        } = this.props;
        this.setState({
            selection: selectedRows,
        }, () => {
            if (onRowSelectionChange) {
                onRowSelectionChange(selectedRows);
            }
        });
    };

    onGetRows = () => {
        const {
            onLoadMoreData,
        } = this.props;
        if (onLoadMoreData) {
            onLoadMoreData();
        }
    };
}

DataTable.propTypes = {
    /** Component used to generate the table cell and its contents */
    cellComponent: PropTypes.oneOfType([
        PropTypes.func,
        PropTypes.object,
    ]),

    /** An array containing column data */
    columns: PropTypes.arrayOf(PropTypes.shape({
        /** The name identifier of the column */
        name: PropTypes.string,

        /** The column header display text */
        title: PropTypes.string,

        /** Overrides the default table column width in pixels or CSS-accepted units (auto, px, %, em, rem, vm, vh, vmin, vmax) */
        width: PropTypes.oneOfType([
            PropTypes.number,
            PropTypes.string,
        ]),
    })).isRequired,

    /** Styles for this component */
    contextStyles: PropTypes.shape({
        /** Styles for the container surrounding the table */
        container: PropTypes.object,
    }),

    /** Default (px) width applied to all columns */
    defaultColumnWidth: PropTypes.number,

    /** Filter entry directives */
    filterDirectives: PropTypes.arrayOf(PropTypes.shape({
        /** The name identifier of the filter column */
        columnName: PropTypes.string,

        /** The filter operation */
        operation: PropTypes.string,

        /** The filter value */
        value: PropTypes.oneOfType([
            PropTypes.string,
            PropTypes.number,
            PropTypes.bool,
        ]),
    })),

    /** Column filter operators */
    filterOperations: PropTypes.arrayOf(PropTypes.shape({
        /** The name identifier of the filter column */
        columnName: PropTypes.string,

        /** The filter operation */
        operators: PropTypes.arrayOf(PropTypes.string),
    })),

    /**
     * Indexes of columns to freeze and affix to the left of the table
     * NOTE: non-contiguous column indexes do NOT reorder columns and dock together as a group.
     * If fixed grouped non-contiguous columns are desired then the columns prop must be reordered with
     * fixed columns at the beginning of the columns array in the desired fixed order.
     */
    fixedColumnIndexes: PropTypes.arrayOf(PropTypes.number),

    /** Enables a scrolling mode in which rows are loaded in sequence */
    infiniteScrolling: PropTypes.bool,

    /** Shows the table loading state when enabled */
    loading: PropTypes.bool,

    /**
     * Called on column freeze toggle component click
     * @param {String} columnName - the name of the column with the freeze toggle
     * @returns {void}
     */
    onColumnFrozen: PropTypes.func,

    /**
     * Called on column reorder
     * @param {Array.<String>} columnsNamesInOrder
     * @returns {void}
     */
    onColumnsReordered: PropTypes.func,

    /**
     * Called on column resize
     * @param {Array.<{ columnName: String, width: Number }>} columnsWithWidths
     * @returns {void}
     */
    onColumnsResized: PropTypes.func,

    /**
     * Called on column filter cancel component click
     * @param {String} columnName - the name identifier of the column
     * @returns {void}
     */
    onFilterCancel: PropTypes.func,

    /**
     * @typedef {Object} Filter
     * @property {String} columnName
     * @property {FilterOperations} operation
     * @property {String} value
     *
     * Called on filter change
     * @param {Array.<Filter>} filters
     * @returns {void}
     */
    onFilterChange: PropTypes.func,

    /**
     * Called on column filters reset component click
     * @param {String} columnName - the name identifier of the column
     * @returns {void}
     */
    onFilterReset: PropTypes.func,

    /**
     * Called on column filters submit component click
     * @param {String} columnName - the name identifier of the column
     * @returns {void}
     */
    onFilterSubmit: PropTypes.func,

    /**
     * Called on column filter toggle component click
     * @param {String} columnName - the name identifier of the column
     * @returns {void}
     */
    onFilterToggle: PropTypes.func,

    /**
     * Called when the viewport row threshold reaches a certain point when scrolling
     * @param {Number} skip
     * @param {Number} take
     * @returns {void}
     */
    onLoadMoreData: PropTypes.func,

    /**
     * Called on row click
     * @param {Array.<Number|String>} selectedRowIndexes
     * @returns {void}
     */
    onRowSelectionChange: PropTypes.func,

    /**
     * Called on column sort ascending toggle component click
     * @param {String} columnName - the name of the column to be sorted
     * @returns {void}
     */
    onSortAscending: PropTypes.func,

    /**
     * Called on column sort clear toggle component click
     * @param {String} columnName - the name of the column to clear sorting
     * @returns {void}
     */
    onSortClear: PropTypes.func,

    /**
     * Called on column sort descending toggle component click
     * @param {String} columnName - the name of the column to be sorted
     * @returns {void}
     */
    onSortDescending: PropTypes.func,

    /** Specifies the count of rows in the current virtual chunk */
    pageSize: PropTypes.number,

    /** An array containing custom data */
    rows: PropTypes.array.isRequired,

    /** Indexes of rows that show as selected in the UI */
    selectedRowIndexes: PropTypes.arrayOf(PropTypes.number),

    /** Shows the table header filter row when enabled */
    showFilterRow: PropTypes.bool,

    /** Column sorting directives */
    sortDirectives: PropTypes.arrayOf(PropTypes.shape({
        /** The name identifier of the column */
        columnName: PropTypes.string,

        /** The sorting direction reflected in the UI */
        direction: PropTypes.oneOf([
            'asc',
            'desc',
            null,
        ]),
    })),

    /** ID string used for testing */
    testID: PropTypes.string,

    /** The total number of rows to render including virtual rows when infinite scrolling is enabled */
    totalRows: PropTypes.number.isRequired,
};

DataTable.defaultProps = {
    cellComponent: DataTableCell,
    contextStyles: {},
    defaultColumnWidth: DEFAULT_COLUMN_WIDTH,
    filterDirectives: [],
    filterOperations: [],
    fixedColumnIndexes: [],
    infiniteScrolling: false,
    selectedRowIndexes: [],
    sortDirectives: [],
    testID: 'data-table-root',
};

export {
    DataTableCell,
    FILTER_OPERATIONS,
    SORT_DIRECTIONS,
};

export default DataTable;
